# Copyright (C) 2006, 2013 Red Hat, Inc.
# Copyright (C) 2006 Daniel P. Berrange <berrange@redhat.com>
#
# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

import glob
import os
import urllib.parse

from gi.repository import Gtk

from virtinst import log

from .lib import uiutil
from .baseclass import vmmGObjectUI
from .connmanager import vmmConnectionManager

(HV_QEMU, HV_XEN, HV_LXC, HV_QEMU_SESSION, HV_BHYVE, HV_VZ, HV_CUSTOM) = range(7)


def _default_uri():  # pragma: no cover
    if os.path.exists("/var/lib/xen"):
        if os.path.exists("/dev/xen/evtchn") or os.path.exists("/proc/xen"):
            return "xen:///"

    if (
        os.path.exists("/usr/bin/qemu")
        or os.path.exists("/usr/bin/qemu-kvm")
        or os.path.exists("/usr/bin/kvm")
        or os.path.exists("/usr/libexec/qemu-kvm")
        or glob.glob("/usr/bin/qemu-system-*")
    ):
        return "qemu:///system"

    if os.path.exists("/usr/lib/libvirt/libvirt_lxc") or os.path.exists(
        "/usr/lib64/libvirt/libvirt_lxc"
    ):
        return "lxc:///"
    return None


class vmmCreateConn(vmmGObjectUI):
    @classmethod
    def get_instance(cls, parentobj):
        try:
            if not cls._instance:
                cls._instance = vmmCreateConn()
            return cls._instance
        except Exception as e:  # pragma: no cover
            parentobj.err.show_err(_("Error launching connect dialog: %s") % str(e))

    def __init__(self):
        vmmGObjectUI.__init__(self, "createconn.ui", "vmm-open-connection")
        self._cleanup_on_app_close()

        self.builder.connect_signals(
            {
                "on_hypervisor_changed": self.hypervisor_changed,
                "on_connect_remote_toggled": self.connect_remote_toggled,
                "on_username_entry_changed": self.username_changed,
                "on_hostname_changed": self.hostname_changed,
                "on_cancel_clicked": self.cancel,
                "on_connect_clicked": self.open_conn,
                "on_vmm_open_connection_delete_event": self.cancel,
            }
        )

        self.set_initial_state()
        self.reset_state()

    @staticmethod
    def default_uri():
        return _default_uri()

    def cancel(self, ignore1=None, ignore2=None):
        log.debug("Cancelling open connection")
        self.close()
        return 1

    def close(self, ignore1=None, ignore2=None):
        log.debug("Closing open connection")
        self.topwin.hide()

    def show(self, parent):
        log.debug("Showing open connection")
        if self.is_visible():
            self.topwin.present()
            return

        self.reset_state()
        self.topwin.set_transient_for(parent)
        self.topwin.present()

    def _cleanup(self):
        pass

    def set_initial_state(self):
        self.widget("connect").grab_default()

        combo = self.widget("hypervisor")
        # [connection ID, label]
        model = Gtk.ListStore(int, str)

        def _add_hv_row(rowid, config_name, label):
            if (
                not self.config.default_hvs
                or not config_name
                or config_name in self.config.default_hvs
            ):
                model.append([rowid, label])

        _add_hv_row(HV_QEMU, "qemu", "QEMU/KVM")
        _add_hv_row(HV_QEMU_SESSION, "qemu", "QEMU/KVM " + _("user session"))
        _add_hv_row(HV_XEN, "xen", "Xen")
        _add_hv_row(HV_LXC, "lxc", "Libvirt-LXC")
        _add_hv_row(HV_BHYVE, "bhyve", "Bhyve")
        _add_hv_row(HV_VZ, "vz", "Virtuozzo")
        _add_hv_row(-1, None, "")
        _add_hv_row(HV_CUSTOM, None, _("Custom URI..."))
        combo.set_model(model)
        uiutil.init_combo_text_column(combo, 1)

        def sepfunc(model, it):
            return model[it][0] == -1

        combo.set_row_separator_func(sepfunc)

    def reset_state(self):
        self.set_default_hypervisor()
        self.widget("autoconnect").set_sensitive(True)
        self.widget("autoconnect").set_active(True)
        self.widget("hostname").set_text("")
        self.widget("connect-remote").set_active(False)
        self.widget("username-entry").set_text("")
        self.widget("uri-entry").set_text("")
        self.connect_remote_toggled(self.widget("connect-remote"))
        self.populate_uri()

    def is_remote(self):
        # Whether user is requesting a remote connection
        return self.widget("connect-remote").get_active()

    def set_default_hypervisor(self):
        default = self.default_uri()
        if not default or default.startswith("qemu"):
            uiutil.set_list_selection(self.widget("hypervisor"), HV_QEMU)
        elif default.startswith("xen"):  # pragma: no cover
            uiutil.set_list_selection(self.widget("hypervisor"), HV_XEN)

    def hostname_changed(self, src_ignore):
        self.populate_uri()

    def hypervisor_changed(self, src):
        ignore = src
        hv = uiutil.get_list_selection(self.widget("hypervisor"))
        is_session = hv == HV_QEMU_SESSION
        is_custom = hv == HV_CUSTOM
        show_remote = not is_session and not is_custom
        uiutil.set_grid_row_visible(self.widget("session-warning-box"), is_session)
        uiutil.set_grid_row_visible(self.widget("connect-remote"), show_remote)
        uiutil.set_grid_row_visible(self.widget("username-entry"), show_remote)
        uiutil.set_grid_row_visible(self.widget("hostname"), show_remote)
        if not show_remote:
            self.widget("connect-remote").set_active(False)

        uiutil.set_grid_row_visible(self.widget("uri-label"), not is_custom)
        uiutil.set_grid_row_visible(self.widget("uri-entry"), is_custom)
        if is_custom:
            label = self.widget("uri-label").get_text()
            self.widget("uri-entry").set_text(label)
            self.widget("uri-entry").grab_focus()
        self.populate_uri()

    def username_changed(self, src_ignore):
        self.populate_uri()

    def connect_remote_toggled(self, src_ignore):
        is_remote = self.is_remote()
        self.widget("hostname").set_sensitive(is_remote)
        self.widget("autoconnect").set_active(not is_remote)
        self.widget("username-entry").set_sensitive(is_remote)

        if is_remote and not self.widget("username-entry").get_text():
            self.widget("username-entry").set_text("root")
        self.populate_uri()

    def populate_uri(self):
        uri = self.generate_uri()
        self.widget("uri-label").set_text(uri)

    def generate_uri(self):
        hv = uiutil.get_list_selection(self.widget("hypervisor"))
        host = self.widget("hostname").get_text().strip()
        user = self.widget("username-entry").get_text()
        is_remote = self.is_remote()

        hvstr = ""
        if hv == HV_XEN:
            hvstr = "xen"
        elif hv == HV_QEMU or hv == HV_QEMU_SESSION:
            hvstr = "qemu"
        elif hv == HV_BHYVE:
            hvstr = "bhyve"
        elif hv == HV_VZ:
            hvstr = "vz"
        else:
            hvstr = "lxc"

        addrstr = ""
        if user:
            addrstr += urllib.parse.quote(user) + "@"

        if host.count(":") > 1:
            host = "[%s]" % host
        addrstr += host

        if is_remote:
            hoststr = "+ssh://" + addrstr + "/"
        else:
            hoststr = ":///"

        uri = hvstr + hoststr
        if hv in (HV_QEMU, HV_BHYVE, HV_VZ):
            uri += "system"
        elif hv == HV_QEMU_SESSION:
            uri += "session"

        return uri

    def validate(self):
        is_remote = self.is_remote()
        host = self.widget("hostname").get_text()

        if is_remote and not host:
            msg = _("A hostname is required for remote connections.")
            return self.err.val_err(msg)

        return True

    def _conn_open_completed(self, conn, ConnectError):
        if not ConnectError:
            self.close()
            self.reset_finish_cursor()
            return

        msg, details, title = ConnectError
        msg += "\n\n"
        msg += _("Would you still like to remember this connection?")

        remember = self.err.show_err(
            msg,
            details,
            title,
            buttons=Gtk.ButtonsType.YES_NO,
            dialog_type=Gtk.MessageType.QUESTION,
            modal=True,
        )
        self.reset_finish_cursor()
        if remember:
            self.close()
        else:
            vmmConnectionManager.get_instance().remove_conn(conn.get_uri())

    def open_conn(self, ignore):
        if not self.validate():
            return

        auto = False
        if self.widget("autoconnect").get_sensitive():
            auto = bool(self.widget("autoconnect").get_active())
        if self.widget("uri-label").is_visible():
            uri = self.generate_uri()
        else:
            uri = self.widget("uri-entry").get_text()

        log.debug("Generate URI=%s, auto=%s", uri, auto)

        conn = vmmConnectionManager.get_instance().add_conn(uri)
        conn.set_autoconnect(auto)
        if conn.is_active():
            self._conn_open_completed(conn, None)
            return

        conn.connect_once("open-completed", self._conn_open_completed)
        self.set_finish_cursor()
        conn.open()



# Copyright (C) 2020 Red Hat, Inc.
#
# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

# This file is a collection of code used for testing
# code paths primarily via our uitests/

import os


def fake_job_info():
    import random

    total = 1024 * 1024 * 1024
    fakepcent = random.choice(range(1, 100))
    remaining = (total / 100) * fakepcent
    return [None, None, None, total, None, remaining]


def fake_interface_addresses(iface, source):
    import libvirt

    mac = iface.macaddr
    if source == libvirt.VIR_DOMAIN_INTERFACE_ADDRESSES_SRC_AGENT:
        ret = {
            "enp1s0": {
                "hwaddr": mac,
                "addrs": [
                    {"addr": "10.0.0.1", "prefix": 24, "type": 0},
                    {"addr": "fd00:beef::1", "prefix": 128, "type": 1},
                    {"addr": "fe80::1", "prefix": 64, "type": 1},
                ],
            },
            "lo": {
                "hwaddr": "00:00:00:00:00:00",
                "addrs": [
                    {"addr": "127.0.0.1", "prefix": 8, "type": 0},
                    {"addr": "::1", "prefix": 128, "type": 1},
                ],
            },
        }
    else:
        ret = {
            "vnet0": {
                "hwaddr": mac,
                "addrs": [{"addr": "10.0.0.3", "prefix": 0, "type": 0}],
            }
        }
    return ret


def fake_dhcp_leases():
    ret = [
        {
            "clientid": "XXX",
            "expirytime": 1598570993,
            "hostname": None,
            "iaid": "1448103320",
            "iface": "virbr1",
            "ipaddr": "fd00:beef::2",
            "mac": "BAD",
            "prefix": 64,
            "type": 1,
        },
        {
            "clientid": "YYY",
            "expirytime": 1598570993,
            "hostname": None,
            "iaid": None,
            "iface": "virbr1",
            "ipaddr": "10.0.0.2",
            "mac": "NOPE",
            "prefix": 24,
            "type": 0,
        },
    ]
    return ret


def schedule_fake_agent_event(conn, cb):
    import libvirt

    vmname = conn.config.CLITestOptions.fake_agent_event
    backend = conn.get_backend()
    state = libvirt.VIR_CONNECT_DOMAIN_EVENT_AGENT_LIFECYCLE_STATE_CONNECTED
    reason = libvirt.VIR_CONNECT_DOMAIN_EVENT_AGENT_LIFECYCLE_REASON_CHANNEL

    def time_cb():
        dom = backend.lookupByName(vmname)
        cb(backend, dom, state, reason, None)

    conn.timeout_add(500, time_cb)


def schedule_fake_nodedev_event(conn, lifecycle_cb, update_cb):
    import libvirt

    nodename = conn.config.CLITestOptions.fake_nodedev_event
    backend = conn.get_backend()

    def lifecycle_cb_wrapper():
        nodedev = backend.nodeDeviceLookupByName(nodename)
        state = libvirt.VIR_NODE_DEVICE_EVENT_CREATED
        reason = 0
        lifecycle_cb(backend, nodedev, state, reason, None)

    def update_cb_wrapper():
        nodedev = backend.nodeDeviceLookupByName(nodename)
        update_cb(backend, nodedev, None)

    conn.timeout_add(500, lifecycle_cb_wrapper)
    conn.timeout_add(1000, update_cb_wrapper)


def fake_openauth(conn, cb, data):
    ignore = conn
    import libvirt

    creds = [
        [libvirt.VIR_CRED_USERNAME, "Username", None, None, None],
        [libvirt.VIR_CRED_PASSPHRASE, "Password", None, None, None],
    ]
    cb(creds, data)
    assert all([bool(cred[4]) for cred in creds])


class fakeVirtBootstrap:
    @staticmethod
    def bootstrap(**kwargs):
        import time
        import logging

        log = logging.getLogger("virtBootstrap")
        log.info("mock virtBootstrap msg1")
        kwargs["progress_cb"]({"status": "msg1"})
        time.sleep(0.5)
        log.info("mock virtBootstrap msg2")
        kwargs["progress_cb"]({"status": "msg2"})
        time.sleep(0.5)
        log.info("mock virtBootstrap msg3")
        kwargs["progress_cb"]({"status": "msg3"})
        if "username" in kwargs:
            raise RuntimeError("fakeVirtBootstrap mock auth failure!")


class CLITestOptionsClass:
    """
    Helper class for parsing and tracking --test-* options.
    The suboptions are:

    * first-run: Run the app with fresh gsettings values saved to
        a keyfile, mimicking a first app run. Also sets
        GSETTINGS to use memory backend, in case any other app
        preferences would be affected. The ui testsuite sets this
        for most tests.

    * xmleditor-enabled: Force the xmleditor gsettings preference on.

    * gsettings-keyfile: Override the gsettings values with those
        from the passed in keyfile, to test with different default
        settings.

    * leak-debug: Enabling this will tell us, at app exit time,
        which vmmGObjects were not garbage collected. This is caused
        by circular references to other objects, like a signal that
        wasn't disconnected. It's not a big deal, but if we have objects
        that can be created and destroyed a lot over the course of
        the app lifecycle, every non-garbage collected class is a
        memory leak. So it's nice to poke at this every now and then
        and try to track down what we need to add to class _cleanup handling.

    * no-events: Force disable libvirt event APIs for testing fallback

    * break_setfacl: For setfacl calls to fail, for test scenarios.
        This is hit via the directory search permissions checking
        for disk image usage for qemu

    * enable-libguestfs: Force enable the libguestfs gsetting
    * disable-libguestfs: Force disable the libguestfs gsetting

    * test-managed-save: Triggers a couple conditions for testing
        managed save issues

    * test-vm-run-fail: Make VM run fail, so we can test the error path
    * test-update-device-fail: Make UpdateDevice API call fail

    * firstrun-uri: If set, use this as the initial connection URI
        if we are doing firstrun testing
    * fake-vnc-username: Fake VNC username auth request
    * fake-systray: Enable the fake systray window
    * fake-virtbootstrap: Mock the virtBootstrap module, since getting
        it to actually work across fedora versions is hard
    * object-denylist=NAME: Make object initialize for that name
        fail to test some connection code paths
    * conn-crash: Test connection abruptly closing like when
        libvirtd is restarted.
    * fake-agent-event: Fake a qemu guest agent API event
    * fake-nodedev-event: Fake nodedev API events
    * fake-openauth: Fake user+pass response from libvirt openauth,
        for testing the TCP URI auth dialog
    * fake-session-error: Fake a connection open error that
        triggers logind session lookup
    * short-poll: Use a polling interval of only .1 seconds to speed
        up the uitests a bit
    """

    def __init__(self, test_options_str):
        optset = set()
        for optstr in test_options_str:
            optset.update(set(optstr.split(",")))

        first_run = self._parse(optset)
        self._process(first_run)

    def _parse(self, optset):
        def _get(optname):
            if optname not in optset:
                return False
            optset.remove(optname)
            return True

        def _get_value(optname):
            for opt in optset:
                if opt.startswith(optname + "="):
                    optset.remove(opt)
                    return opt.split("=", 1)[1]

        first_run = _get("first-run")
        self.leak_debug = _get("leak-debug")
        self.no_events = _get("no-events")
        self.xmleditor_enabled = _get("xmleditor-enabled")
        self.gsettings_keyfile = _get_value("gsettings-keyfile")
        self.break_setfacl = _get("break-setfacl")
        self.disable_libguestfs = _get("disable-libguestfs")
        self.enable_libguestfs = _get("enable-libguestfs")
        self.test_managed_save = _get("test-managed-save")
        self.test_vm_run_fail = _get("test-vm-run-fail")
        self.test_update_device_fail = _get("test-update-device-fail")
        self.firstrun_uri = _get_value("firstrun-uri")
        self.fake_vnc_username = _get("fake-vnc-username")
        self.fake_systray = _get("fake-systray")
        self.object_denylist = _get_value("object-denylist")
        self.conn_crash = _get("conn-crash")
        self.fake_agent_event = _get_value("fake-agent-event")
        self.fake_nodedev_event = _get_value("fake-nodedev-event")
        self.fake_openauth = _get("fake-openauth")
        self.fake_session_error = _get("fake-session-error")
        self.short_poll = _get("short-poll")
        self.fake_virtbootstrap = _get("fake-virtbootstrap")

        if optset:  # pragma: no cover
            raise RuntimeError("Unknown --test-options keys: %s" % optset)

        return first_run

    def _process(self, first_run):
        if first_run:
            # So other settings like gtk are reset and not affected
            os.environ["GSETTINGS_BACKEND"] = "memory"

        if first_run and not self.gsettings_keyfile:
            import atexit
            import tempfile

            filename = tempfile.mktemp(prefix="virtmanager-firstrun-keyfile")
            self.gsettings_keyfile = filename
            atexit.register(lambda: os.unlink(filename))

        if self.break_setfacl:
            import virtinst.diskbackend

            def fake_search(*args, **kwargs):
                raise RuntimeError("Fake search fix fail from test suite")

            virtinst.diskbackend.SETFACL = "getfacl"
            # pylint: disable=protected-access
            virtinst.diskbackend._fix_perms_chmod = fake_search



# Copyright (C) 2014 Red Hat, Inc.
#
# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

from gi.repository import Gtk

import virtinst
from virtinst import log

from ..lib import uiutil
from ..baseclass import vmmGObjectUI


NET_ROW_LABEL = 0
NET_ROW_SENSITIVE = 1
NET_ROW_DATA = 2


class _NetRowData:
    @staticmethod
    def build_row(*args, **kwargs):
        rowdata = _NetRowData(*args, **kwargs)
        return [rowdata.label, rowdata.is_sensitive, rowdata]

    def __init__(self, nettype, source, label, is_sensitive, manual=False):
        self.nettype = nettype
        self.source = source
        self.label = label
        self.is_sensitive = is_sensitive
        self.manual = manual


def _build_manual_row(nettype, label):
    return _NetRowData.build_row(nettype, None, label, True, manual=True)


def _pretty_network_desc(nettype, source=None, netobj=None):
    if nettype == virtinst.DeviceInterface.TYPE_USER:
        return _("Usermode networking")

    extra = None
    if nettype == virtinst.DeviceInterface.TYPE_VIRTUAL:
        ret = _("Virtual network")
        if netobj:
            extra = ": %s" % netobj.pretty_forward_mode()
    else:
        ret = nettype.capitalize()

    if source:
        ret += " '%s'" % source
    if extra:
        ret += " %s" % extra

    return ret


class vmmNetworkList(vmmGObjectUI):
    __gsignals__ = {
        "changed": (vmmGObjectUI.RUN_FIRST, None, []),
    }

    def __init__(self, conn, builder, topwin):
        vmmGObjectUI.__init__(self, "netlist.ui", None, builder=builder, topwin=topwin)
        self.conn = conn

        self.builder.connect_signals(
            {
                "on_net_source_changed": self._on_net_source_changed,
                "on_net_portgroup_changed": self._emit_changed,
                "on_net_bridge_name_changed": self._emit_changed,
            }
        )

        self._init_ui()
        self.top_label = self.widget("net-source-label")
        self.top_box = self.widget("net-source-box")

    def _cleanup(self):
        self.conn.disconnect_by_obj(self)
        self.conn = None

        self.top_label.destroy()
        self.top_box.destroy()

    ##########################
    # Initialization methods #
    ##########################

    def _init_ui(self):
        fields = []
        fields.insert(NET_ROW_LABEL, str)
        fields.insert(NET_ROW_SENSITIVE, bool)
        fields.insert(NET_ROW_DATA, object)

        model = Gtk.ListStore(*fields)
        combo = self.widget("net-source")
        combo.set_model(model)

        text = Gtk.CellRendererText()
        combo.pack_start(text, True)
        combo.add_attribute(text, "text", NET_ROW_LABEL)
        combo.add_attribute(text, "sensitive", NET_ROW_SENSITIVE)

        combo = self.widget("net-portgroup")
        model = Gtk.ListStore(str)
        combo.set_model(model)
        uiutil.init_combo_text_column(combo, 0)

        self.conn.connect("net-added", self._repopulate_network_list)
        self.conn.connect("net-removed", self._repopulate_network_list)

    def _find_virtual_networks(self):
        rows = []

        for net in self.conn.list_nets():
            nettype = virtinst.DeviceInterface.TYPE_VIRTUAL

            label = _pretty_network_desc(nettype, net.get_name(), net)
            if not net.is_active():
                label += " (%s)" % _("Inactive")

            if net.get_xmlobj().virtualport_type == "openvswitch":
                label += " (OpenVSwitch)"

            row = _NetRowData.build_row(nettype, net.get_name(), label, True)
            rows.append(row)

        return rows

    def _populate_network_model(self, model):
        model.clear()

        def _add_manual_bridge_row():
            _nettype = virtinst.DeviceInterface.TYPE_BRIDGE
            _label = _("Bridge device...")
            model.append(_build_manual_row(_nettype, _label))
            return len(model) - 1

        def _add_manual_macvtap_row():
            _label = _("Macvtap device...")
            _nettype = virtinst.DeviceInterface.TYPE_DIRECT
            model.append(_build_manual_row(_nettype, _label))

        def _add_manual_vdpa_row():
            _label = _("vDPA device...")
            _nettype = virtinst.DeviceInterface.TYPE_VDPA
            model.append(_build_manual_row(_nettype, _label))

        vnets = self._find_virtual_networks()
        default_bridge = virtinst.DeviceInterface.default_bridge(self.conn.get_backend())

        add_usermode = False
        if self.conn.is_qemu_unprivileged():
            log.debug("Using unprivileged qemu, adding usermode net")
            default_bridge = None
            add_usermode = True

        if add_usermode:
            nettype = virtinst.DeviceInterface.TYPE_USER
            label = _pretty_network_desc(nettype)
            model.append(_NetRowData.build_row(nettype, None, label, True))

        defaultnetidx = None
        for row in sorted(vnets, key=lambda r: r[NET_ROW_LABEL]):
            model.append(row)
            if row[NET_ROW_DATA].source == "default":
                defaultnetidx = len(model) - 1

        bridgeidx = _add_manual_bridge_row()
        _add_manual_macvtap_row()
        _add_manual_vdpa_row()

        # If there is a bridge device, default to that
        if default_bridge:
            self.widget("net-manual-source").set_text(default_bridge)
            return bridgeidx

        # If not, use 'default' network
        if defaultnetidx is not None:
            return defaultnetidx

        # If not present, use first list entry
        if bridgeidx == 0:
            # This means we are defaulting to something that
            # requires manual intervention. Raise the warning
            self.widget("net-default-warn-box").show()
        return 0

    def _check_network_is_running(self, net):
        # Make sure VirtualNetwork is running
        if not net.type == virtinst.DeviceInterface.TYPE_VIRTUAL:
            return
        devname = net.source

        netobj = None
        if net.type == virtinst.DeviceInterface.TYPE_VIRTUAL:
            netobj = self.conn.get_net_by_name(devname)

        if not netobj or netobj.is_active():
            return

        res = self.err.yes_no(
            _("Virtual Network is not active."),
            _("Virtual Network '%s' is not active. Would you like to start the network now?")
            % devname,
        )
        if not res:
            return  # pragma: no cover

        # Try to start the network
        try:
            netobj.start()
            log.debug("Started network '%s'", devname)
        except Exception as e:  # pragma: no cover
            return self.err.show_err(
                _("Could not start virtual network '%(device)s': %(error)s")
                % {
                    "device": devname,
                    "error": str(e),
                }
            )

    def _find_rowiter_for_dev(self, net):
        """
        Find the row in our current model that matches the passed in
        net device (like populating the details UI for an existing VM).
        If we don't find a match, we fake it a bit
        """
        nettype = net.type
        source = net.source
        if net.network:
            # If using type=network with a forward mode=bridge network,
            # on domain startup the runtime XML will be changed to
            # type=bridge and both source/@bridge and source/@network will
            # be filled in. For our purposes, treat this as a type=network
            source = net.network
            nettype = "network"

        combo = self.widget("net-source")

        def _find_row(_nettype, _source, _manual):
            for row in combo.get_model():
                rowdata = row[NET_ROW_DATA]
                if _nettype and rowdata.nettype != _nettype:
                    continue
                if _source and rowdata.source != _source:
                    continue
                if _manual and rowdata.manual != _manual:
                    continue  # pragma: no cover
                return row.iter

        # Find the matching row in the net list
        rowiter = _find_row(nettype, source, None)
        if rowiter:
            return rowiter

        # If this is a bridge or macvtap device, show the
        # manual source mode
        if nettype in [
            virtinst.DeviceInterface.TYPE_BRIDGE,
            virtinst.DeviceInterface.TYPE_DIRECT,
            virtinst.DeviceInterface.TYPE_VDPA,
        ]:
            rowiter = _find_row(nettype, None, True)
            self.widget("net-manual-source").set_text(source or "")
            if rowiter:
                return rowiter

        # This is some network type we don't know about. Generate
        # a label for it and stuff it in the list
        desc = _pretty_network_desc(nettype, source)
        combo.get_model().insert(0, _NetRowData.build_row(nettype, source, desc, True))
        return combo.get_model()[0].iter

    ###############
    # Public APIs #
    ###############

    def _get_network_row_data(self):
        return uiutil.get_list_selection(self.widget("net-source"), column=NET_ROW_DATA)

    def get_network_selection(self):
        rowdata = self._get_network_row_data()
        net_type = rowdata.nettype
        net_src = rowdata.source
        net_check_manual = rowdata.manual

        if net_check_manual:
            net_src = self.widget("net-manual-source").get_text() or None

        mode = None
        is_direct = net_type == virtinst.DeviceInterface.TYPE_DIRECT
        if is_direct:
            # This is generally the safest and most featureful default
            mode = "bridge"

        portgroup = None
        if self.widget("net-portgroup").is_visible():
            portgroup = uiutil.get_list_selection(self.widget("net-portgroup"))

        return net_type, net_src, mode, portgroup

    def build_device(self, macaddr, model=None):
        nettype, devname, mode, portgroup = self.get_network_selection()

        net = virtinst.DeviceInterface(self.conn.get_backend())
        net.type = nettype
        net.source = devname
        net.macaddr = macaddr
        net.model = model
        net.source_mode = mode
        net.portgroup = portgroup

        return net

    def validate_device(self, net):
        self._check_network_is_running(net)
        virtinst.DeviceInterface.check_mac_in_use(net.conn, net.macaddr)
        net.validate()

    def reset_state(self):
        self.widget("net-default-warn-box").set_visible(False)
        self.widget("net-manual-source").set_text("")
        self.widget("net-portgroup").get_child().set_text("")
        self._repopulate_network_list()

    def set_dev(self, net):
        self.reset_state()
        rowiter = self._find_rowiter_for_dev(net)

        combo = self.widget("net-source")
        combo.set_active_iter(rowiter)
        combo.emit("changed")

        if net.portgroup:
            uiutil.set_list_selection(self.widget("net-portgroup"), net.portgroup)

    #############
    # Listeners #
    #############

    def _emit_changed(self, *args, **kwargs):
        ignore1 = args
        ignore2 = kwargs
        self.emit("changed")

    def _repopulate_network_list(self, *args, **kwargs):
        ignore1 = args
        ignore2 = kwargs

        netlist = self.widget("net-source")
        current_label = uiutil.get_list_selection(netlist, column=NET_ROW_LABEL)

        model = netlist.get_model()
        if not model:
            return  # pragma: no cover

        try:
            if model:
                netlist.set_model(None)
                default_idx = self._populate_network_model(model)
        finally:
            netlist.set_model(model)

        for row in netlist.get_model():
            if current_label and row[NET_ROW_LABEL] == current_label:
                netlist.set_active_iter(row.iter)
                return

        netlist.set_active(default_idx)

    def _populate_portgroups(self, portgroups):
        combo = self.widget("net-portgroup")
        model = combo.get_model()
        model.clear()

        default = None
        for p in portgroups:
            model.append([p.name])
            if p.default:
                default = p.name

        uiutil.set_list_selection(combo, default)

    def _on_net_source_changed(self, src):
        ignore = src
        self._emit_changed()
        rowdata = self._get_network_row_data()
        if not rowdata:
            return  # pragma: no cover

        nettype = rowdata.nettype
        is_direct = nettype == virtinst.DeviceInterface.TYPE_DIRECT
        is_virtual = nettype == virtinst.DeviceInterface.TYPE_VIRTUAL

        uiutil.set_grid_row_visible(self.widget("net-macvtap-warn-box"), is_direct)

        show_bridge = rowdata.manual
        uiutil.set_grid_row_visible(self.widget("net-manual-source"), show_bridge)

        net = None
        if is_virtual:
            net = self.conn.get_net_by_name(rowdata.source)

        portgroups = []
        if net:
            portgroups = net.get_xmlobj().portgroups

        uiutil.set_grid_row_visible(self.widget("net-portgroup"), bool(portgroups))
        self._populate_portgroups(portgroups)



# Copyright (C) 2007, 2013-2014 Red Hat, Inc.
# Copyright (C) 2007 Daniel P. Berrange <berrange@redhat.com>
#
# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

from gi.repository import Gtk
from gi.repository import Pango

from virtinst import log

from .lib import uiutil
from .asyncjob import vmmAsyncJob
from .baseclass import vmmGObjectUI
from .createnet import vmmCreateNetwork
from .xmleditor import vmmXMLEditor


EDIT_NET_IDS = (
    EDIT_NET_NAME,
    EDIT_NET_AUTOSTART,
    EDIT_NET_XML,
) = list(range(3))


ICON_RUNNING = "state_running"
ICON_SHUTOFF = "state_shutoff"


class vmmHostNets(vmmGObjectUI):
    def __init__(self, conn, builder, topwin):
        vmmGObjectUI.__init__(self, "hostnets.ui", None, builder=builder, topwin=topwin)
        self.conn = conn

        self._addnet = None
        self._xmleditor = None

        self._active_edits = set()
        self.top_box = self.widget("top-box")

        self.builder.connect_signals(
            {
                "on_net_add_clicked": self._add_network_cb,
                "on_net_delete_clicked": self._delete_network_cb,
                "on_net_stop_clicked": self._stop_network_cb,
                "on_net_start_clicked": self._start_network_cb,
                "on_net_apply_clicked": (lambda *x: self._net_apply()),
                "on_net_list_changed": self._net_selected_cb,
                "on_net_autostart_toggled": (lambda *x: self._enable_net_apply(EDIT_NET_AUTOSTART)),
                "on_net_name_changed": (lambda *x: self._enable_net_apply(EDIT_NET_NAME)),
            }
        )

        self._init_ui()
        self._populate_networks()
        self._refresh_conn_state()
        self.conn.connect("net-added", self._conn_nets_changed_cb)
        self.conn.connect("net-removed", self._conn_nets_changed_cb)
        self.conn.connect("state-changed", self._conn_state_changed_cb)

    #######################
    # Standard UI methods #
    #######################

    def _cleanup(self):
        self.conn = None

        if self._addnet:
            self._addnet.cleanup()
            self._addnet = None

        self._xmleditor.cleanup()
        self._xmleditor = None

    def close(self, ignore1=None, ignore2=None):
        if self._addnet:
            self._addnet.close()

    ###########
    # UI init #
    ###########

    def _init_ui(self):
        self.widget("network-pages").set_show_tabs(False)

        self._xmleditor = vmmXMLEditor(
            self.builder, self.topwin, self.widget("net-details-align"), self.widget("net-details")
        )
        self._xmleditor.connect("changed", lambda s: self._enable_net_apply(EDIT_NET_XML))
        self._xmleditor.connect("xml-requested", self._xmleditor_xml_requested_cb)
        self._xmleditor.connect("xml-reset", self._xmleditor_xml_reset_cb)

        # [ netobj, label, icon name, icon size, is_active ]
        netListModel = Gtk.ListStore(object, str, str, int, bool)
        self.widget("net-list").set_model(netListModel)

        sel = self.widget("net-list").get_selection()
        sel.set_select_function((lambda *x: self._confirm_changes()), None)

        netCol = Gtk.TreeViewColumn(_("Networks"))
        netCol.set_spacing(6)
        net_txt = Gtk.CellRendererText()
        net_txt.set_property("ellipsize", Pango.EllipsizeMode.END)
        net_img = Gtk.CellRendererPixbuf()
        netCol.pack_start(net_img, False)
        netCol.pack_start(net_txt, True)
        netCol.add_attribute(net_txt, "text", 1)
        netCol.add_attribute(net_txt, "sensitive", 4)
        netCol.add_attribute(net_img, "icon-name", 2)
        netCol.add_attribute(net_img, "stock-size", 3)
        self.widget("net-list").append_column(netCol)
        netListModel.set_sort_column_id(1, Gtk.SortType.ASCENDING)

    ##############
    # Public API #
    ##############

    def refresh_page(self):
        self.conn.schedule_priority_tick(pollnet=True)

    #################
    # UI populating #
    #################

    def _refresh_conn_state(self):
        conn_active = self.conn.is_active()
        self.widget("net-add").set_sensitive(conn_active and self.conn.support.conn_network())

        if conn_active and not self.conn.support.conn_network():
            self._set_error_page(  # pragma: no cover
                _("Libvirt connection does not support virtual network management.")
            )

        if conn_active:
            uiutil.set_list_selection_by_number(self.widget("net-list"), 0)
            return

        self._populate_networks()
        self._set_error_page(_("Connection not active."))

    def _current_network(self):
        return uiutil.get_list_selection(self.widget("net-list"))

    def _set_error_page(self, msg):
        self.widget("network-pages").set_current_page(1)
        self.widget("network-error-label").set_text(msg)
        self.widget("net-start").set_sensitive(False)
        self.widget("net-stop").set_sensitive(False)
        self.widget("net-delete").set_sensitive(False)
        self._disable_net_apply()

    def _refresh_current_network(self):
        net = self._current_network()
        if not net:
            self._set_error_page(_("No virtual network selected."))
            return

        self.widget("network-pages").set_current_page(0)

        try:
            self._populate_net_state(net)
        except Exception as e:  # pragma: no cover
            log.exception(e)
            self._set_error_page(_("Error selecting network: %s") % e)
        self._disable_net_apply()

    def _populate_networks(self):
        net_list = self.widget("net-list")
        curnet = self._current_network()

        model = net_list.get_model()
        # Prevent events while the model is modified
        net_list.set_model(None)
        try:
            net_list.get_selection().unselect_all()
            model.clear()
            for net in self.conn.list_nets():
                net.disconnect_by_obj(self)
                net.connect("state-changed", self._net_state_changed_cb)
                model.append(
                    [
                        net,
                        net.get_name(),
                        "network-idle",
                        Gtk.IconSize.LARGE_TOOLBAR,
                        bool(net.is_active()),
                    ]
                )
        finally:
            net_list.set_model(model)

        uiutil.set_list_selection(net_list, curnet)

    def _populate_net_ipv4_state(self, net):
        (netstr, (dhcpstart, dhcpend)) = net.get_ipv4_network()

        self.widget("net-ipv4-expander").set_visible(bool(netstr))
        if not netstr:
            return

        self.widget("net-ipv4-forwarding").set_text(net.pretty_forward_mode())

        dhcpstr = _("Disabled")
        if dhcpstart:
            dhcpstr = dhcpstart + " - " + dhcpend
        self.widget("net-ipv4-dhcp-range").set_text(dhcpstr)
        self.widget("net-ipv4-network").set_text(netstr)

    def _populate_net_ipv6_state(self, net):
        (netstr, (dhcpstart, dhcpend)) = net.get_ipv6_network()

        self.widget("net-ipv6-expander").set_visible(bool(netstr))

        if netstr:
            prettymode = _("Routed network")
        elif net.get_ipv6_enabled():
            prettymode = _("Isolated network, internal routing only")
        else:
            prettymode = _("Isolated network, routing disabled")
        self.widget("net-ipv6-forwarding").set_text(prettymode)

        dhcpstr = _("Disabled")
        if dhcpstart:
            dhcpstr = dhcpstart + " - " + dhcpend
        self.widget("net-ipv6-dhcp-range").set_text(dhcpstr)
        self.widget("net-ipv6-network").set_text(netstr or "")

    def _populate_net_state(self, net):
        active = net.is_active()

        self.widget("net-details").set_sensitive(True)
        self.widget("net-name").set_text(net.get_name())
        self.widget("net-name").set_editable(not active)
        self.widget("net-device").set_text(net.get_bridge_device() or "")
        self.widget("net-name-domain").set_text(net.get_name_domain() or "")
        uiutil.set_grid_row_visible(self.widget("net-name-domain"), bool(net.get_name_domain()))

        icon = active and ICON_RUNNING or ICON_SHUTOFF
        self.widget("net-state").set_text(net.run_status())
        self.widget("net-state-icon").set_from_icon_name(icon, Gtk.IconSize.BUTTON)

        self.widget("net-start").set_sensitive(not active)
        self.widget("net-stop").set_sensitive(active)
        self.widget("net-delete").set_sensitive(not active)

        autostart = net.get_autostart()
        self.widget("net-autostart").set_active(autostart)
        self.widget("net-autostart").set_label(_("On Boot"))

        self._populate_net_ipv4_state(net)
        self._populate_net_ipv6_state(net)

        self._xmleditor.set_xml_from_libvirtobject(net)

    #############################
    # Network lifecycle actions #
    #############################

    def _delete_network_cb(self, src):
        net = self._current_network()
        if net is None:
            return  # pragma: no cover

        result = self.err.yes_no(
            _("Are you sure you want to permanently delete the network %s?") % net.get_name()
        )
        if not result:
            return

        log.debug("Deleting network '%s'", net.get_name())
        vmmAsyncJob.simple_async_noshow(
            net.delete, [], self, _("Error deleting network '%s'") % net.get_name()
        )

    def _start_network_cb(self, src):
        net = self._current_network()
        if net is None:
            return  # pragma: no cover

        log.debug("Starting network '%s'", net.get_name())
        vmmAsyncJob.simple_async_noshow(
            net.start, [], self, _("Error starting network '%s'") % net.get_name()
        )

    def _stop_network_cb(self, src):
        net = self._current_network()
        if net is None:
            return  # pragma: no cover

        log.debug("Stopping network '%s'", net.get_name())
        vmmAsyncJob.simple_async_noshow(
            net.stop, [], self, _("Error stopping network '%s'") % net.get_name()
        )

    def _add_network_cb(self, src):
        log.debug("Launching 'Add Network'")
        try:
            if self._addnet is None:
                self._addnet = vmmCreateNetwork(self.conn)
            self._addnet.show(self.topwin)
        except Exception as e:  # pragma: no cover
            self.err.show_err(_("Error launching network wizard: %s") % str(e))

    ############################
    # Net apply/config actions #
    ############################

    def _net_apply(self):
        net = self._current_network()
        if net is None:
            return  # pragma: no cover

        log.debug("Applying changes for network '%s'", net.get_name())
        try:
            if EDIT_NET_AUTOSTART in self._active_edits:
                auto = self.widget("net-autostart").get_active()
                net.set_autostart(auto)
            if EDIT_NET_NAME in self._active_edits:
                net.define_name(self.widget("net-name").get_text())
                self.idle_add(self._populate_networks)
            if EDIT_NET_XML in self._active_edits:
                net.define_xml(self._xmleditor.get_xml())

        except Exception as e:
            self.err.show_err(_("Error changing network settings: %s") % str(e))
            return
        finally:
            self._disable_net_apply()

    def _disable_net_apply(self):
        self._active_edits = set()
        self.widget("net-apply").set_sensitive(False)
        self._xmleditor.details_changed = False

    def _enable_net_apply(self, edittype):
        self.widget("net-apply").set_sensitive(True)
        self._active_edits.add(edittype)
        self._xmleditor.details_changed = True

    def _confirm_changes(self):
        if self.is_visible() and self._active_edits and self.err.confirm_unapplied_changes():
            self._net_apply()

        self._disable_net_apply()
        return True

    ################
    # UI listeners #
    ################

    def _conn_state_changed_cb(self, conn):
        self._refresh_conn_state()

    def _conn_nets_changed_cb(self, src, net):
        self._populate_networks()

    def _net_state_changed_cb(self, net):
        # Update net state inline in the tree model
        for row in self.widget("net-list").get_model():
            if row[0] == net:
                row[4] = net.is_active()

        # If refreshed network is the current net, refresh the UI
        curnet = self._current_network()
        if curnet == net:
            self._refresh_current_network()

    def _net_selected_cb(self, selection):
        self._refresh_current_network()

    def _xmleditor_xml_requested_cb(self, src):
        self._refresh_current_network()

    def _xmleditor_xml_reset_cb(self, src):
        self._refresh_current_network()



# Copyright (C) 2009, 2013, 2014 Red Hat, Inc.
# Copyright (C) 2009 Cole Robinson <crobinso@redhat.com>
#
# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

import os

from gi.repository import Gtk

from virtinst import log

from .lib import uiutil
from .baseclass import vmmGObjectUI
from .hoststorage import vmmHostStorage


class _BrowseReasonMetadata:
    def __init__(self, browse_reason):
        self.enable_create = False
        self.storage_title = None
        self.local_title = None
        self.gsettings_key = None
        self.dialog_type = None
        self.choose_label = None

        if browse_reason == vmmStorageBrowser.REASON_IMAGE:
            self.enable_create = True
            self.local_title = _("Locate existing storage")
            self.storage_title = _("Locate or create storage volume")
            self.dialog_type = Gtk.FileChooserAction.SAVE
            self.choose_label = _("_Open")
            self.gsettings_key = "image"

        if browse_reason == vmmStorageBrowser.REASON_ISO_MEDIA:
            self.local_title = _("Locate ISO media")
            self.storage_title = _("Locate ISO media volume")
            self.gsettings_key = "media"

        if browse_reason == vmmStorageBrowser.REASON_FLOPPY_MEDIA:
            self.local_title = _("Locate floppy media")
            self.storage_title = _("Locate floppy media volume")
            self.gsettings_key = "media"

        if browse_reason == vmmStorageBrowser.REASON_FS:
            self.local_title = _("Locate directory volume")
            self.storage_title = _("Locate directory volume")
            self.dialog_type = Gtk.FileChooserAction.SELECT_FOLDER

        if browse_reason is None:
            self.enable_create = True
            self.storage_title = _("Choose Storage Volume")


class vmmStorageBrowser(vmmGObjectUI):
    REASON_IMAGE = "image"
    REASON_ISO_MEDIA = "isomedia"
    REASON_FLOPPY_MEDIA = "floppymedia"
    REASON_FS = "fs"

    def __init__(self, conn):
        vmmGObjectUI.__init__(self, "storagebrowse.ui", "vmm-storage-browse")
        self.conn = conn

        self._first_run = False
        self._finish_cb = None
        self._browse_reason = None

        self.storagelist = vmmHostStorage(
            self.conn, self.builder, self.topwin, self._vol_sensitive_cb
        )
        self._init_ui()

        self.builder.connect_signals(
            {
                "on_vmm_storage_browse_delete_event": self.close,
            }
        )
        self.bind_escape_key_close()

    def show(self, parent):
        log.debug("Showing storage browser")
        if not self._first_run:
            self._first_run = True
            pool = self.conn.get_default_pool()
            uiutil.set_list_selection(self.storagelist.widget("pool-list"), pool)

        self.topwin.set_transient_for(parent)
        self.topwin.present()
        self.conn.schedule_priority_tick(pollpool=True)

    def close(self, ignore1=None, ignore2=None):
        if self.is_visible():
            log.debug("Closing storage browser")
            self.topwin.hide()
        self.storagelist.close()
        return 1

    def _cleanup(self):
        self.conn = None

        self.storagelist.cleanup()
        self.storagelist = None

    ###########
    # UI init #
    ###########

    def _init_ui(self):
        self.storagelist.connect("browse-clicked", self._browse_clicked)
        self.storagelist.connect("volume-chosen", self._volume_chosen)
        self.storagelist.connect("cancel-clicked", self.close)

        self.widget("storage-align").add(self.storagelist.top_box)
        self.err.set_modal_default(True)
        self.storagelist.err.set_modal_default(True)

        tooltip = ""
        is_remote = self.conn.is_remote()
        self.storagelist.widget("browse-local").set_sensitive(not is_remote)
        if is_remote:
            tooltip = _("Cannot use local storage on remote connection.")
        self.storagelist.widget("browse-local").set_tooltip_text(tooltip)

        uiutil.set_grid_row_visible(self.storagelist.widget("pool-autostart"), False)
        uiutil.set_grid_row_visible(self.storagelist.widget("pool-name-entry"), False)
        uiutil.set_grid_row_visible(self.storagelist.widget("pool-state-box"), False)
        self.storagelist.widget("browse-local").set_visible(True)
        self.storagelist.widget("browse-cancel").set_visible(True)
        self.storagelist.widget("choose-volume").set_visible(True)
        self.storagelist.widget("choose-volume").set_sensitive(False)
        self.storagelist.widget("pool-apply").set_visible(False)

        self.set_browse_reason(self._browse_reason)

    ##############
    # Public API #
    ##############

    def set_finish_cb(self, callback):
        self._finish_cb = callback

    def set_vm_name(self, name):
        self.storagelist.set_name_hint(name)

    def set_browse_reason(self, reason):
        self._browse_reason = reason
        data = _BrowseReasonMetadata(self._browse_reason)

        self.topwin.set_title(data.storage_title)
        self.storagelist.widget("vol-add").set_sensitive(data.enable_create)

    #############
    # Listeners #
    #############

    def _browse_clicked(self, src):
        ignore = src
        return self._browse_local()

    def _volume_chosen(self, src, volume):
        ignore = src
        log.debug("Chosen volume XML:\n%s", volume.xmlobj.get_xml())
        self._finish(volume.get_target_path())

    def _vol_sensitive_cb(self, fmt):
        if (self._browse_reason == vmmStorageBrowser.REASON_FS) and fmt != "dir":
            return False
        return True

    ####################
    # Internal helpers #
    ####################

    def _browse_local(self):
        data = _BrowseReasonMetadata(self._browse_reason)
        gsettings_key = data.gsettings_key

        start_folder = None
        if gsettings_key:
            start_folder = self.config.get_default_directory(gsettings_key)

        filename = self.err.browse_local(
            dialog_type=data.dialog_type,
            dialog_name=data.local_title,
            start_folder=start_folder,
            choose_label=data.choose_label,
        )

        if not filename:
            return

        log.debug("Browse local chose path=%s", filename)

        if gsettings_key:
            self.config.set_default_directory(gsettings_key, os.path.dirname(filename))

        self._finish(filename)

    def _finish(self, path):
        if self._finish_cb:
            self._finish_cb(self, path)
        self.close()



# Copyright (C) 2009, 2013 Red Hat, Inc.
# Copyright (C) 2009 Cole Robinson <crobinso@redhat.com>
#
# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

import traceback

from gi.repository import Gtk
from gi.repository import Pango

from virtinst import log
from virtinst import xmlutil

from .lib import uiutil
from .asyncjob import vmmAsyncJob
from .baseclass import vmmGObjectUI
from .connmanager import vmmConnectionManager
from .object.domain import vmmDomain
from .xmleditor import vmmXMLEditor


NUM_COLS = 3
(COL_LABEL, COL_URI, COL_CAN_MIGRATE) = range(NUM_COLS)


class vmmMigrateDialog(vmmGObjectUI):
    @classmethod
    def show_instance(cls, parentobj, vm):
        try:
            if not cls._instance:
                cls._instance = vmmMigrateDialog()
            cls._instance.show(parentobj.topwin, vm)
        except Exception as e:  # pragma: no cover
            parentobj.err.show_err(_("Error launching migrate dialog: %s") % str(e))

    def __init__(self):
        vmmGObjectUI.__init__(self, "migrate.ui", "vmm-migrate")
        self.vm = None

        self._xmleditor = vmmXMLEditor(
            self.builder, self.topwin, self.widget("details-box-align"), self.widget("details-box")
        )
        self._xmleditor.connect("xml-requested", self._xmleditor_xml_requested_cb)

        self.builder.connect_signals(
            {
                "on_vmm_migrate_delete_event": self._delete_event,
                "on_migrate_cancel_clicked": self._cancel_clicked,
                "on_migrate_finish_clicked": self._finish_clicked,
                "on_migrate_dest_changed": self._destconn_changed,
                "on_migrate_set_address_toggled": self._set_address_toggled,
                "on_migrate_set_port_toggled": self._set_port_toggled,
                "on_migrate_mode_changed": self._mode_changed,
            }
        )
        self.bind_escape_key_close()
        self._cleanup_on_app_close()

        self._init_state()

    def _cleanup(self):
        self.vm = None
        self._xmleditor.cleanup()
        self._xmleditor = None

    @property
    def _connobjs(self):
        return vmmConnectionManager.get_instance().conns

    @property
    def conn(self):
        return self.vm and self.vm.conn or None

    ##############
    # Public API #
    ##############

    def show(self, parent, vm):
        log.debug("Showing migrate wizard")
        self._set_vm(vm)
        self._reset_state()
        self.topwin.set_transient_for(parent)
        self.topwin.present()

    def close(self, ignore1=None, ignore2=None):
        log.debug("Closing migrate wizard")
        self.topwin.hide()
        self._set_vm(None)
        return 1

    def _vm_removed_cb(self, _conn, vm):
        if self.vm == vm:
            self.close()

    def _set_vm(self, newvm):
        oldvm = self.vm
        if oldvm:
            oldvm.conn.disconnect_by_obj(self)
        if newvm:
            newvm.conn.connect("vm-removed", self._vm_removed_cb)
        self.vm = newvm

    ################
    # Init helpers #
    ################

    def _init_state(self):
        # Connection combo
        cols = [None] * NUM_COLS
        cols[COL_LABEL] = str
        cols[COL_URI] = str
        cols[COL_CAN_MIGRATE] = bool
        model = Gtk.ListStore(*cols)
        combo = self.widget("migrate-dest")
        combo.set_model(model)
        text = uiutil.init_combo_text_column(combo, COL_LABEL)
        text.set_property("ellipsize", Pango.EllipsizeMode.MIDDLE)
        text.set_property("width-chars", 30)
        combo.add_attribute(text, "sensitive", COL_CAN_MIGRATE)
        model.set_sort_column_id(COL_LABEL, Gtk.SortType.ASCENDING)

        def _sorter(model, iter1, iter2, ignore):
            def _cmp(a, b):
                return (a > b) - (a < b)

            row1 = model[iter1]
            row2 = model[iter2]
            if row1[COL_URI] is None:
                return -1
            return _cmp(row1[COL_LABEL], row2[COL_LABEL])

        model.set_sort_func(COL_LABEL, _sorter)

        # Mode combo
        combo = self.widget("migrate-mode")
        # label, is_tunnel
        model = Gtk.ListStore(str, bool)
        model.append([_("Direct"), False])
        model.append([_("Tunnelled"), True])
        combo.set_model(model)
        uiutil.init_combo_text_column(combo, 0)

        self.widget("migrate-dest").emit("changed")

        self.widget("migrate-mode").set_tooltip_text(
            self.widget("migrate-mode-label").get_tooltip_text()
        )
        self.widget("migrate-unsafe").set_tooltip_text(
            self.widget("migrate-unsafe-label").get_tooltip_text()
        )
        self.widget("migrate-temporary").set_tooltip_text(
            self.widget("migrate-temporary-label").get_tooltip_text()
        )

    def _reset_state(self):
        self._xmleditor.reset_state()

        title_str = _("<span size='large'>Migrate '%(vm)s'</span>") % {
            "vm": xmlutil.xml_escape(self.vm.get_name()),
        }
        self.widget("header-label").set_markup(title_str)

        self.widget("migrate-advanced-expander").set_expanded(False)

        self.widget("migrate-cancel").grab_focus()

        self.widget("config-box").set_visible(True)

        hostname = self.conn.libvirt_gethostname()
        srctext = "%s (%s)" % (hostname, self.conn.get_pretty_desc())
        self.widget("migrate-label-name").set_text(self.vm.get_name_or_title())
        self.widget("migrate-label-src").set_text(srctext)
        self.widget("migrate-label-src").set_tooltip_text(self.conn.get_uri())

        self.widget("migrate-set-address").set_active(True)
        self.widget("migrate-set-address").emit("toggled")
        self.widget("migrate-set-port").set_active(True)

        self.widget("migrate-mode").set_active(0)
        self.widget("migrate-unsafe").set_active(False)
        self.widget("migrate-temporary").set_active(False)

        if self.conn.is_xen():
            # Default xen port is 8002
            self.widget("migrate-port").set_value(8002)
        else:
            # QEMU migrate port range is 49152+64
            self.widget("migrate-port").set_value(49152)

        self._populate_destconn()

    #############
    # Listeners #
    #############

    def _delete_event(self, ignore1, ignore2):
        self.close()
        return 1

    def _cancel_clicked(self, src):
        ignore = src
        self.close()

    def _finish_clicked(self, src):
        ignore = src
        self._finish()

    def _destconn_changed(self, src):
        row = uiutil.get_list_selected_row(src)
        if not row:
            return

        can_migrate = row and row[COL_CAN_MIGRATE] or False
        uri = row[COL_URI]

        tooltip = ""
        if not can_migrate:
            tooltip = _("A valid destination connection must be selected.")

        self.widget("config-box").set_visible(can_migrate)
        self.widget("migrate-finish").set_sensitive(can_migrate)
        self.widget("migrate-finish").set_tooltip_text(tooltip)

        address = ""
        tunnel_warning = ""
        tunnel_uri = ""

        if can_migrate and uri in self._connobjs:
            destconn = self._connobjs[uri]

            tunnel_uri = destconn.get_uri()
            if not destconn.is_remote():
                tunnel_warning = _(
                    "A remotely accessible libvirt URI "
                    "is required for tunneled migration, but the "
                    "selected connection is a local URI. Libvirt will "
                    "reject this unless you add a transport."
                )
                tunnel_warning = "<span size='small'>%s</span>" % tunnel_warning

            address = destconn.libvirt_gethostname()

        self.widget("migrate-address").set_text(address)

        self.widget("migrate-tunnel-uri").set_text(tunnel_uri)
        uiutil.set_grid_row_visible(self.widget("migrate-tunnel-warning-box"), bool(tunnel_warning))
        self.widget("migrate-tunnel-warning-label").set_markup(tunnel_warning)

    def _set_address_toggled(self, src):
        enable = src.get_active()
        self.widget("migrate-address").set_visible(enable)
        self.widget("migrate-address-label").set_visible(not enable)

        port_enable = self.widget("migrate-set-port").get_active()
        self.widget("migrate-set-port").set_active(enable and port_enable)
        self.widget("migrate-set-port").emit("toggled")

    def _set_port_toggled(self, src):
        enable = src.get_active()
        self.widget("migrate-port").set_visible(enable)
        self.widget("migrate-port-label").set_visible(not enable)

    def _is_tunnel_selected(self):
        return uiutil.get_list_selection(self.widget("migrate-mode"), column=1)

    def _mode_changed(self, src):
        ignore = src
        is_tunnel = self._is_tunnel_selected()
        self.widget("migrate-direct-box").set_visible(not is_tunnel)
        self.widget("migrate-tunnel-box").set_visible(is_tunnel)

    ###########################
    # destconn combo handling #
    ###########################

    def _build_dest_row(self, destconn):
        driver = self.conn.get_driver()
        origuri = self.conn.get_uri()

        can_migrate = False
        pretty_uri = destconn.get_pretty_desc()
        desc = pretty_uri
        desturi = destconn.get_uri()

        if destconn.get_driver() != driver:
            desc = _("%(uri)s (Hypervisors do not match)") % {"uri": pretty_uri}
        elif destconn.is_disconnected():
            desc = _("%(uri)s (Disconnected)") % {"uri": pretty_uri}
        elif destconn.get_uri() == origuri:
            desc = _("%(uri)s (Same connection)") % {"uri": pretty_uri}
        elif destconn.is_active():
            can_migrate = True

        return [desc, desturi, can_migrate]

    def _populate_destconn(self):
        combo = self.widget("migrate-dest")
        model = combo.get_model()
        model.clear()

        rows = []
        for conn in list(self._connobjs.values()):
            rows.append(self._build_dest_row(conn))

        if not any([row[COL_CAN_MIGRATE] for row in rows]):
            rows.insert(0, [_("No usable connections available."), None, False])

        for row in rows:
            model.append(row)

        combo.set_active(0)
        for idx, row in enumerate(model):
            if row[COL_CAN_MIGRATE]:
                combo.set_active(idx)
                break

    ####################
    # migrate handling #
    ####################

    def _build_regular_migrate_uri(self):
        address = None
        if self.widget("migrate-address").get_visible():
            address = self.widget("migrate-address").get_text()

        port = None
        if self.widget("migrate-port").get_visible():
            port = int(self.widget("migrate-port").get_value())

        if not address:
            return

        if self.conn.is_xen():
            uri = "%s" % address
        else:
            uri = "tcp:%s" % address
        if port:
            uri += ":%s" % port
        return uri

    def _finish_cb(self, error, details, destconn):
        self.reset_finish_cursor()

        if error:
            error = _("Unable to migrate guest: %s") % error
            self.err.show_err(error, details=details)
        else:
            destconn.schedule_priority_tick(pollvm=True)
            self.conn.schedule_priority_tick(pollvm=True)
            self.close()

    def _finish(self):
        try:
            xml = None
            if self._xmleditor.is_xml_selected():
                xml = self._xmleditor.get_xml()
                log.debug("Using XML from xmleditor:\n%s", xml)

            row = uiutil.get_list_selected_row(self.widget("migrate-dest"))
            destlabel = row[COL_LABEL]
            destconn = self._connobjs.get(row[COL_URI])

            tunnel = self._is_tunnel_selected()
            unsafe = self.widget("migrate-unsafe").get_active()
            temporary = self.widget("migrate-temporary").get_active()

            if tunnel:
                uri = self.widget("migrate-tunnel-uri").get_text()
            else:
                uri = self._build_regular_migrate_uri()
        except Exception as e:  # pragma: no cover
            details = "".join(traceback.format_exc())
            self.err.show_err((_("Uncaught error validating input: %s") % str(e)), details=details)
            return

        self.set_finish_cursor()

        cancel_cb = None
        if self.vm.supports_domain_job_info():
            cancel_cb = (self._cancel_migration, self.vm)

        if uri:
            destlabel += " " + uri

        progWin = vmmAsyncJob(
            self._async_migrate,
            [self.vm, destconn, uri, tunnel, unsafe, temporary, xml],
            self._finish_cb,
            [destconn],
            _("Migrating VM '%s'") % self.vm.get_name(),
            (
                _("Migrating VM '%(name)s' to %(host)s. This may take a while.")
                % {"name": self.vm.get_name(), "host": destlabel}
            ),
            self.topwin,
            cancel_cb=cancel_cb,
        )
        progWin.run()

    def _cancel_migration(self, asyncjob, vm):
        log.debug("Cancelling migrate job")
        try:
            vm.abort_job()
        except Exception as e:
            log.exception("Error cancelling migrate job")
            asyncjob.show_warning(_("Error cancelling migrate job: %s") % e)
            return

        asyncjob.job_canceled = True  # pragma: no cover

    def _async_migrate(
        self, asyncjob, origvm, origdconn, migrate_uri, tunnel, unsafe, temporary, xml
    ):
        meter = asyncjob.get_meter()

        srcconn = origvm.conn
        dstconn = origdconn

        vminst = srcconn.get_backend().lookupByName(origvm.get_name())
        vm = vmmDomain(srcconn, vminst, vminst.UUID())

        log.debug(
            "Migrating vm=%s from %s to %s", vm.get_name(), srcconn.get_uri(), dstconn.get_uri()
        )

        vm.migrate(dstconn, migrate_uri, tunnel, unsafe, temporary, xml, meter=meter)

    ################
    # UI listeners #
    ################

    def _xmleditor_xml_requested_cb(self, src):
        self._xmleditor.set_xml(self.vm.xmlobj.get_xml())



# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

from virtinst import DeviceTpm

from ..lib import uiutil
from ..baseclass import vmmGObjectUI


def _tpm_pretty_model(val):
    labels = {
        DeviceTpm.MODEL_TIS: _("TIS"),
        DeviceTpm.MODEL_CRB: _("CRB"),
        DeviceTpm.MODEL_SPAPR: _("SPAPR"),
    }
    return labels.get(val, val)


_EDIT_TPM_ENUM = range(1, 5)
(
    _EDIT_TPM_TYPE,
    _EDIT_TPM_DEVICE_PATH,
    _EDIT_TPM_MODEL,
    _EDIT_TPM_VERSION,
) = _EDIT_TPM_ENUM


class vmmTPMDetails(vmmGObjectUI):
    __gsignals__ = {
        "changed": (vmmGObjectUI.RUN_FIRST, None, []),
    }

    def __init__(self, vm, builder, topwin):
        super().__init__("tpmdetails.ui", None, builder=builder, topwin=topwin)
        self.vm = vm
        self.conn = vm.conn

        self._active_edits = []

        def _e(edittype):
            def signal_cb(*args):
                self._change_cb(edittype)

            return signal_cb

        self.builder.connect_signals(
            {
                "on_tpm_type_changed": _e(_EDIT_TPM_TYPE),
                "on_tpm_device_path_changed": _e(_EDIT_TPM_DEVICE_PATH),
                "on_tpm_model_changed": _e(_EDIT_TPM_MODEL),
                "on_tpm_version_changed": _e(_EDIT_TPM_VERSION),
            }
        )

        self._init_ui()
        self.top_box = self.widget("top-box")

    def _cleanup(self):
        self.vm = None
        self.conn = None

    ##############
    # UI helpers #
    ##############

    def _init_ui(self):
        domcaps = self.vm.get_domain_capabilities()

        # We could check domcaps for this, but emulated is really the
        # preferred default here, so just let it fail
        rows = [
            [DeviceTpm.TYPE_EMULATOR, _("Emulated")],
            [DeviceTpm.TYPE_PASSTHROUGH, _("Passthrough")],
        ]
        uiutil.build_simple_combo(self.widget("tpm-type"), rows, sort=False)

        rows = [[None, _("Hypervisor default")]]
        if domcaps.devices.tpm.present:
            values = domcaps.devices.tpm.get_enum("model").get_values()
        else:
            values = [DeviceTpm.MODEL_CRB, DeviceTpm.MODEL_TIS]
        for v in values:
            rows.append([v, _tpm_pretty_model(v)])

        uiutil.build_simple_combo(self.widget("tpm-model"), rows, sort=False)

        rows = [
            [None, _("Hypervisor default")],
            [DeviceTpm.VERSION_2_0, DeviceTpm.VERSION_2_0],
            [DeviceTpm.VERSION_1_2, DeviceTpm.VERSION_1_2],
        ]
        uiutil.build_simple_combo(self.widget("tpm-version"), rows, sort=False)

    def _sync_ui(self):
        devtype = uiutil.get_list_selection(self.widget("tpm-type"))

        uiutil.set_grid_row_visible(
            self.widget("tpm-device-path"), devtype == DeviceTpm.TYPE_PASSTHROUGH
        )
        uiutil.set_grid_row_visible(self.widget("tpm-version"), devtype == DeviceTpm.TYPE_EMULATOR)

    ##################
    # Public UI APIs #
    ##################

    def reset_state(self):
        self.widget("tpm-device-path").set_text("/dev/tpm0")
        uiutil.set_list_selection(self.widget("tpm-type"), DeviceTpm.TYPE_EMULATOR)

        default_model = DeviceTpm.default_model(self.vm.xmlobj)
        uiutil.set_list_selection(self.widget("tpm-model"), default_model)
        uiutil.set_list_selection(self.widget("tpm-version"), None)

    def set_dev(self, dev):
        self.reset_state()

        uiutil.set_list_selection(self.widget("tpm-type"), dev.type)
        uiutil.set_list_selection(self.widget("tpm-model"), dev.model)
        uiutil.set_list_selection(self.widget("tpm-version"), dev.version)
        self.widget("tpm-device-path").set_text(dev.device_path or "")

        self._active_edits = []

    ########################
    # Device building APIs #
    ########################

    def _set_values(self, dev):
        typ = uiutil.get_list_selection(self.widget("tpm-type"))
        model = uiutil.get_list_selection(self.widget("tpm-model"))
        device_path = self.widget("tpm-device-path").get_text()
        version = uiutil.get_list_selection(self.widget("tpm-version"))

        if not self.widget("tpm-device-path").get_visible():
            device_path = None
        if not self.widget("tpm-version").get_visible():
            version = None

        if _EDIT_TPM_TYPE in self._active_edits:
            dev.type = typ
        if _EDIT_TPM_MODEL in self._active_edits:
            dev.model = model
        if _EDIT_TPM_DEVICE_PATH in self._active_edits:
            dev.device_path = device_path
        if _EDIT_TPM_VERSION in self._active_edits:
            dev.version = version

        return dev

    def build_device(self):
        self._active_edits = _EDIT_TPM_ENUM[:]

        conn = self.conn.get_backend()
        dev = DeviceTpm(conn)
        self._set_values(dev)

        dev.validate()
        return dev

    def update_device(self, dev):
        newdev = DeviceTpm(dev.conn, parsexml=dev.get_xml())
        self._set_values(newdev)
        return newdev

    #############
    # Listeners #
    #############

    def _change_cb(self, edittype):
        self._sync_ui()
        if edittype not in self._active_edits:
            self._active_edits.append(edittype)
        self.emit("changed")



# Copyright (C) 2006, 2013 Red Hat, Inc.
# Copyright (C) 2006 Hugh O. Brock <hbrock@redhat.com>
#
# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

import threading
import traceback

from gi.repository import GLib

import libvirt

import virtinst.progress

from .baseclass import vmmGObjectUI


class _vmmMeter(virtinst.progress.Meter):
    def __init__(self, pbar_pulse, pbar_fraction, pbar_done):
        virtinst.progress.Meter.__init__(self, quiet=True)

        self._pbar_pulse = pbar_pulse
        self._pbar_fraction = pbar_fraction
        self._pbar_done = pbar_done

    #################
    # Internal APIs #
    #################

    def _write(self):
        if self._size is None:
            self._pbar_pulse("", self._text)
        else:
            fread = virtinst.progress.Meter.format_number(self._total_read)
            rtime = virtinst.progress.Meter.format_time(self._meter.re.remaining_time(), True)
            frac = self._meter.re.fraction_read()
            out = "%3i%% %5sB %s ETA" % (frac * 100, fread, rtime)
            self._pbar_fraction(frac, out, self._text)

    #############################################
    # Public APIs specific to virt-manager code #
    #############################################

    def is_started(self):
        return bool(self._meter.start_time)

    ###################
    # Meter overrides #
    ###################

    def start(self, *args, **kwargs):
        super().start(*args, **kwargs)
        self._write()

    def update(self, *args, **kwargs):
        super().update(*args, **kwargs)
        self._write()

    def end(self, *args, **kwargs):
        super().end(*args, **kwargs)
        self._pbar_done()


def cb_wrapper(callback, asyncjob, *args, **kwargs):
    try:
        callback(asyncjob, *args, **kwargs)
    except Exception as e:
        # If job is cancelled, don't report error to user.
        if isinstance(e, libvirt.libvirtError) and asyncjob.can_cancel() and asyncjob.job_canceled:
            return  # pragma: no cover

        asyncjob.set_error(str(e), "".join(traceback.format_exc()))


def _simple_async_done_cb(error, details, parent, errorintro, errorcb, finish_cb):
    if error:
        if errorcb:
            errorcb(error, details)
        else:
            error = errorintro + ": " + error
            parent.err.show_err(error, details=details)

    if finish_cb:
        finish_cb()


def _simple_async(
    callback, args, parent, title, text, errorintro, show_progress, simplecb, errorcb, finish_cb
):
    """
    @show_progress: Whether to actually show a progress dialog
    @simplecb: If true, build a callback wrapper that ignores the asyncjob
               param that's passed to every cb by default
    """
    docb = callback
    if simplecb:

        def tmpcb(job, *args, **kwargs):
            ignore = job
            callback(*args, **kwargs)

        docb = tmpcb

    asyncjob = vmmAsyncJob(
        docb,
        args,
        _simple_async_done_cb,
        (parent, errorintro, errorcb, finish_cb),
        title,
        text,
        parent.topwin,
        show_progress=show_progress,
    )
    asyncjob.run()


def idle_wrapper(fn):
    def wrapped(self, *args, **kwargs):
        return self.idle_add(fn, self, *args, **kwargs)

    return wrapped


class vmmAsyncJob(vmmGObjectUI):
    """
    Displays a progress bar while executing the "callback" method.
    """

    @staticmethod
    def simple_async(
        callback, args, parent, title, text, errorintro, simplecb=True, errorcb=None, finish_cb=None
    ):
        _simple_async(
            callback, args, parent, title, text, errorintro, True, simplecb, errorcb, finish_cb
        )

    @staticmethod
    def simple_async_noshow(
        callback, args, parent, errorintro, simplecb=True, errorcb=None, finish_cb=None
    ):
        _simple_async(
            callback, args, parent, "", "", errorintro, False, simplecb, errorcb, finish_cb
        )

    def __init__(
        self,
        callback,
        args,
        finish_cb,
        finish_args,
        title,
        text,
        parent,
        show_progress=True,
        cancel_cb=None,
    ):
        """
        @show_progress: If False, don't actually show a progress dialog
        @cancel_cb: Cancel callback if operation supports it.
            (cb, arg1, arg2, ...)
        """
        vmmGObjectUI.__init__(self, "asyncjob.ui", "vmm-progress")
        self.topwin.set_transient_for(parent)

        self.show_progress = bool(show_progress)

        cancel_cb = cancel_cb or (None, [])
        self.cancel_cb = cancel_cb[0]
        self.cancel_args = [self] + list(cancel_cb[1:])
        self.job_canceled = False
        self._finish_cb = finish_cb
        self._finish_args = finish_args or ()

        self._timer = None
        self._error_info = None
        self._data = None

        self._details_widget = None
        self._details_update_cb = None

        self._is_pulsing = True
        self._meter = None

        self._bg_thread = threading.Thread(target=cb_wrapper, args=[callback, self] + args)
        self._bg_thread.daemon = True

        self.builder.connect_signals(
            {
                "on_async_job_cancel_clicked": self._on_cancel,
            }
        )

        # UI state
        self.topwin.set_title(title)
        self.widget("pbar-text").set_text(text)
        self.widget("cancel-async-job").set_visible(bool(self.cancel_cb))

    ####################
    # Internal helpers #
    ####################

    def _cleanup(self):
        self._bg_thread = None
        self.cancel_cb = None
        self.cancel_args = None
        self._meter = None

    def _set_stage_text(self, text, canceling=False):
        # This should be thread safe, since it's only ever called from
        # pbar idle callbacks and cancel routine which is invoked from the
        # main thread
        if self.job_canceled and not canceling:
            return  # pragma: no cover
        self.widget("pbar-stage").set_text(text)

    ################
    # UI listeners #
    ################

    def _on_cancel(self, ignore1=None, ignore2=None):
        if not self.cancel_cb or not self._bg_thread.is_alive():
            return  # pragma: no cover

        self.cancel_cb(*self.cancel_args)
        if self.job_canceled:  # pragma: no cover
            self.widget("warning-box").hide()
            self._set_stage_text(_("Cancelling job..."), canceling=True)

    ##############
    # Public API #
    ##############

    def get_meter(self):
        if not self._meter:
            self._meter = _vmmMeter(self._pbar_pulse, self._pbar_fraction, self._pbar_done)
        return self._meter

    def set_error(self, error, details):
        self._error_info = (error, details)

    def has_error(self):
        return bool(self._error_info)

    def can_cancel(self):
        return bool(self.cancel_cb)

    def show_warning(self, summary):
        # This should only be called from cancel callbacks, not a the thread
        markup = "<small>%s</small>" % summary
        self.widget("warning-box").show()
        self.widget("warning-text").set_markup(markup)

    def _thread_finished(self):
        GLib.source_remove(self._timer)
        self.topwin.destroy()
        self.cleanup()

        error = None
        details = None
        if self._error_info:
            # pylint: disable=unpacking-non-sequence
            error, details = self._error_info
        self._finish_cb(error, details, *self._finish_args)

    def run(self):
        self._timer = GLib.timeout_add(100, self._exit_if_necessary)

        if self.show_progress:
            self.topwin.present()

        if not self.cancel_cb and self.show_progress:
            self._set_cursor("progress")
        self._bg_thread.start()

    ####################################################################
    # All functions after this point are called from the timer loop or #
    # the worker thread, so anything that touches Gtk needs to be      #
    # dispatches with idle_add                                         #
    ####################################################################

    def _exit_if_necessary(self):
        if not self._bg_thread.is_alive():
            self._thread_finished()
            return False

        if not self._is_pulsing or not self.show_progress:
            return True

        self._pbar_do_pulse()
        return True

    @idle_wrapper
    def _pbar_do_pulse(self):
        if not self.builder:
            return  # pragma: no cover
        self.widget("pbar").pulse()

    @idle_wrapper
    def _pbar_pulse(self, progress="", stage=None):
        self._is_pulsing = True
        if not self.builder:
            return  # pragma: no cover
        self.widget("pbar").set_text(progress)
        self._set_stage_text(stage or _("Processing..."))

    @idle_wrapper
    def _pbar_fraction(self, frac, progress, stage=None):
        self._is_pulsing = False
        if not self.builder:
            return  # pragma: no cover
        self._set_stage_text(stage or _("Processing..."))
        self.widget("pbar").set_text(progress)

        frac = min(frac, 1)
        frac = max(frac, 0)
        self.widget("pbar").set_fraction(frac)

    @idle_wrapper
    def _pbar_done(self):
        self._is_pulsing = False

    @idle_wrapper
    def details_enable(self):
        from gi.repository import Vte

        self._details_widget = Vte.Terminal()
        self.widget("details-box").add(self._details_widget)
        self._details_widget.set_visible(True)
        self.widget("details").set_visible(True)

    @idle_wrapper
    def details_update(self, data):
        self._details_widget.feed(data.replace("\n", "\r\n").encode())



# Copyright (C) 2006, 2013 Red Hat, Inc.
# Copyright (C) 2006 Daniel P. Berrange <berrange@redhat.com>
#
# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

# pylint: disable=wrong-import-order,ungrouped-imports
import gi
from gi.repository import Gdk
from gi.repository import Gtk

from virtinst import log

# We can use either 2.91 or 2.90. This is just to silence runtime warnings
try:
    gi.require_version("Vte", "2.91")
    log.debug("Using VTE API 2.91")
except ValueError:  # pragma: no cover
    gi.require_version("Vte", "2.90")
    log.debug("Using VTE API 2.90")
from gi.repository import Vte

import libvirt

from ..baseclass import vmmGObject


class _DataStream(vmmGObject):
    """
    Wrapper class for interacting with libvirt console stream
    """

    def __init__(self, vm):
        vmmGObject.__init__(self)

        self.vm = vm
        self.conn = vm.conn

        self._stream = None

        self._streamToTerminal = b""
        self._terminalToStream = ""

    def _cleanup(self):
        self.close()

        self.vm = None
        self.conn = None

    #################
    # Internal APIs #
    #################

    def _display_data(self, terminal):
        if not self._streamToTerminal:
            return  # pragma: no cover

        terminal.feed(self._streamToTerminal)
        self._streamToTerminal = b""

    def _event_on_stream(self, stream, events, opaque):
        ignore = stream
        terminal = opaque

        if (
            events & libvirt.VIR_EVENT_HANDLE_ERROR or events & libvirt.VIR_EVENT_HANDLE_HANGUP
        ):  # pragma: no cover
            log.debug("Received stream ERROR/HANGUP, closing console")
            self.close()
            return

        if events & libvirt.VIR_EVENT_HANDLE_READABLE:
            try:
                got = self._stream.recv(1024 * 100)
            except Exception:  # pragma: no cover
                log.exception("Error receiving stream data")
                self.close()
                return

            if got == -2:  # pragma: no cover
                # This is basically EAGAIN
                return
            if len(got) == 0:
                log.debug("Received EOF from stream, closing")
                self.close()
                return

            queued_text = bool(self._streamToTerminal)
            self._streamToTerminal += got
            if not queued_text:
                self.idle_add(self._display_data, terminal)

        if events & libvirt.VIR_EVENT_HANDLE_WRITABLE and self._terminalToStream:

            try:
                done = self._stream.send(self._terminalToStream.encode())
            except Exception:  # pragma: no cover
                log.exception("Error sending stream data")
                self.close()
                return

            if done == -2:  # pragma: no cover
                # This is basically EAGAIN
                return

            self._terminalToStream = self._terminalToStream[done:]

        if not self._terminalToStream:
            self._stream.eventUpdateCallback(
                libvirt.VIR_STREAM_EVENT_READABLE
                | libvirt.VIR_STREAM_EVENT_ERROR
                | libvirt.VIR_STREAM_EVENT_HANGUP
            )

    ##############
    # Public API #
    ##############

    def open(self, dev, terminal):
        if self._stream:
            return

        name = dev and dev.alias.name or None
        log.debug("Opening console stream for dev=%s alias=%s", dev, name)
        # libxl doesn't set aliases, their open_console just defaults to
        # opening the first console device, so don't force presence of
        # an alias

        stream = self.conn.get_backend().newStream(libvirt.VIR_STREAM_NONBLOCK)
        self.vm.open_console(name, stream)
        self._stream = stream

        self._stream.eventAddCallback(
            (
                libvirt.VIR_STREAM_EVENT_READABLE
                | libvirt.VIR_STREAM_EVENT_ERROR
                | libvirt.VIR_STREAM_EVENT_HANGUP
            ),
            self._event_on_stream,
            terminal,
        )

    def close(self):
        if self._stream:
            try:
                self._stream.eventRemoveCallback()
            except Exception:  # pragma: no cover
                log.exception("Error removing stream callback")
            try:
                self._stream.finish()
            except Exception:  # pragma: no cover
                log.exception("Error finishing stream")

        self._stream = None

    def send_data(self, src, text, length, terminal):
        """
        Callback when data has been entered into VTE terminal
        """
        ignore = src
        ignore = length
        ignore = terminal

        if self._stream is None:
            return  # pragma: no cover

        self._terminalToStream += text
        if self._terminalToStream:
            self._stream.eventUpdateCallback(
                libvirt.VIR_STREAM_EVENT_READABLE
                | libvirt.VIR_STREAM_EVENT_WRITABLE
                | libvirt.VIR_STREAM_EVENT_ERROR
                | libvirt.VIR_STREAM_EVENT_HANGUP
            )


class vmmSerialConsole(vmmGObject):
    @staticmethod
    def can_connect(_vm, dev):
        """
        Check if we think we can actually open passed console/serial dev
        """
        usable_types = ["pty"]
        ctype = dev.type

        err = ""

        if ctype not in usable_types:
            err = _("Console for device type '%s' is not supported") % ctype

        return err

    @staticmethod
    def get_serialcon_devices(vm):
        serials = vm.xmlobj.devices.serial
        consoles = vm.xmlobj.devices.console
        if serials and vm.serial_is_console_dup(serials[0]):
            consoles.pop(0)
        return serials + consoles

    def __init__(self, vm, target_port, name):
        vmmGObject.__init__(self)

        self.vm = vm
        self.target_port = target_port
        self.name = name
        self.lastpath = None

        self._datastream = _DataStream(self.vm)

        self._serial_popup = None
        self._serial_copy = None
        self._serial_paste = None
        self._init_popup()

        self._vteterminal = None
        self._init_terminal()

        self._box = None
        self._error_label = None
        self._init_ui()

        self.vm.connect("state-changed", self._vm_status_changed)

    def _cleanup(self):
        self._datastream.cleanup()
        self._datastream = None

        self.vm = None
        self._vteterminal = None
        self._box = None

    ###########
    # UI init #
    ###########

    def _init_terminal(self):
        self._vteterminal = Vte.Terminal()
        self._vteterminal.set_scrollback_lines(1000)
        self._vteterminal.set_audible_bell(False)
        self._vteterminal.get_accessible().set_name("Serial Terminal")

        self._vteterminal.connect("button-press-event", self._show_serial_rcpopup)
        self._vteterminal.connect("commit", self._datastream.send_data, self._vteterminal)
        self._vteterminal.show()

    def _init_popup(self):
        self._serial_popup = Gtk.Menu()
        self._serial_popup.get_accessible().set_name("serial-popup-menu")

        self._serial_copy = Gtk.MenuItem.new_with_mnemonic(_("_Copy"))
        self._serial_copy.connect("activate", self._serial_copy_text)
        self._serial_popup.add(self._serial_copy)

        self._serial_paste = Gtk.MenuItem.new_with_mnemonic(_("_Paste"))
        self._serial_paste.connect("activate", self._serial_paste_text)
        self._serial_popup.add(self._serial_paste)

    def _init_ui(self):
        self._box = Gtk.Notebook()
        self._box.set_show_tabs(False)
        self._box.set_show_border(False)

        align = Gtk.Box()
        align.set_border_width(2)
        evbox = Gtk.EventBox()
        evbox.modify_bg(Gtk.StateType.NORMAL, Gdk.Color(0, 0, 0))
        terminalbox = Gtk.HBox()
        scrollbar = Gtk.VScrollbar()
        self._error_label = Gtk.Label()
        self._error_label.set_width_chars(40)
        self._error_label.set_line_wrap(True)

        if self._vteterminal:
            scrollbar.set_adjustment(self._vteterminal.get_vadjustment())
            align.pack_start(self._vteterminal, True, True, 0)

        evbox.add(align)
        terminalbox.pack_start(evbox, True, True, 0)
        terminalbox.pack_start(scrollbar, False, False, 0)

        self._box.append_page(terminalbox, Gtk.Label(""))
        self._box.append_page(self._error_label, Gtk.Label(""))
        self._box.show_all()

        scrollbar.hide()
        scrollbar.get_adjustment().connect("changed", self._scrollbar_adjustment_changed, scrollbar)

    ###################
    # Private methods #
    ###################

    def _show_error(self, msg):
        self._error_label.set_markup("<b>%s</b>" % msg)
        self._box.set_current_page(1)

    def _lookup_dev(self):
        devs = vmmSerialConsole.get_serialcon_devices(self.vm)
        found = None
        for dev in devs:
            port = dev.get_xml_idx()
            path = dev.source.path

            if port == self.target_port:
                if path != self.lastpath:
                    log.debug("Serial console '%s' path changed to %s", self.target_port, path)
                self.lastpath = path
                found = dev
                break

        if not found:  # pragma: no cover
            log.debug("No devices found for serial target port '%s'", self.target_port)
            self.lastpath = None
        return found

    ##############
    # Public API #
    ##############

    def close(self):
        if self._datastream:
            self._datastream.close()

    def get_box(self):
        return self._box

    def has_focus(self):
        return bool(self._vteterminal and self._vteterminal.get_property("has-focus"))

    def set_focus_callbacks(self, in_cb, out_cb):
        self._vteterminal.connect("focus-in-event", in_cb)
        self._vteterminal.connect("focus-out-event", out_cb)

    def open_console(self):
        try:
            dev = self._lookup_dev()
            self._datastream.open(dev, self._vteterminal)
            self._box.set_current_page(0)
            return True
        except Exception as e:
            log.exception("Error opening serial console")
            self._show_error(_("Error connecting to text console: %s") % e)
            try:
                self._datastream.close()
            except Exception:  # pragma: no cover
                pass
        return False

    ################
    # UI listeners #
    ################

    def _vm_status_changed(self, vm):
        if vm.status() in [libvirt.VIR_DOMAIN_RUNNING]:
            self.open_console()
        else:
            self._datastream.close()

    def _scrollbar_adjustment_changed(self, adjustment, scrollbar):
        scrollbar.set_visible(adjustment.get_upper() > adjustment.get_page_size())

    def _show_serial_rcpopup(self, src, event):
        if event.button != 3:
            return

        self._serial_popup.show_all()

        if src.get_has_selection():
            self._serial_copy.set_sensitive(True)
        else:
            self._serial_copy.set_sensitive(False)
        self._serial_popup.popup_at_pointer(event)

    def _serial_copy_text(self, src_ignore):
        self._vteterminal.copy_clipboard()

    def _serial_paste_text(self, src_ignore):
        self._vteterminal.paste_clipboard()



# Copyright (C) 2011, 2013 Red Hat, Inc.
#
# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

import queue
import threading

from virtinst import log

from ..baseclass import vmmGObject
from ..connmanager import vmmConnectionManager
from ..object.domain import vmmInspectionApplication, vmmInspectionData


def _inspection_error(_errstr):
    data = vmmInspectionData()
    data.errorstr = _errstr
    return data


def _make_fake_data(vm):
    """
    Return fake vmmInspectionData for use with the test driver
    """
    if not vm.xmlobj.devices.disk:
        return _inspection_error("Fake test error no disks")

    data = vmmInspectionData()
    data.os_type = "test_os_type"
    data.distro = "test_distro"
    data.major_version = 123
    data.minor_version = 456
    data.hostname = "test_hostname"
    data.product_name = "test_product_name"
    data.product_variant = "test_product_variant"

    from gi.repository import Gtk

    icontheme = Gtk.IconTheme.get_default()
    icon = icontheme.lookup_icon("vm_new", Gtk.IconSize.LARGE_TOOLBAR, 0)
    data.icon = open(icon.get_filename(), "rb").read()

    data.applications = []
    for prefix in ["test_app1_", "test_app2_"]:
        import time

        app = vmmInspectionApplication()
        if "app1" in prefix:
            app.display_name = prefix + "display_name"
            app.summary = prefix + "summary-" + str(time.time())
        else:
            app.name = prefix + "name"
            app.description = prefix + "description-" + str(time.time()) + "\n"
        app.epoch = 1
        app.version = "2"
        app.release = "3"
        data.applications.append(app)

    return data


def _perform_inspection(conn, vm):  # pragma: no cover
    """
    Perform the actual guestfs interaction and return results in
    a vmmInspectionData object
    """
    import guestfs  # pylint: disable=import-error

    g = guestfs.GuestFS(close_on_exit=False, python_return_dict=True)
    prettyvm = conn.get_uri() + ":" + vm.get_name()
    try:
        g.add_libvirt_dom(vm.get_backend(), readonly=1)
        g.launch()
    except Exception as e:
        log.debug("%s: Error launching libguestfs appliance: %s", prettyvm, str(e))
        return _inspection_error(_("Error launching libguestfs appliance: %s") % str(e))

    log.debug("%s: inspection appliance connected", prettyvm)

    # Inspect the operating system.
    roots = g.inspect_os()
    if len(roots) == 0:
        log.debug("%s: no operating systems found", prettyvm)
        return _inspection_error(_("Inspection found no operating systems."))

    # Arbitrarily pick the first root device.
    root = roots[0]

    # Inspection results.
    os_type = g.inspect_get_type(root)  # eg. "linux"
    distro = g.inspect_get_distro(root)  # eg. "fedora"
    major_version = g.inspect_get_major_version(root)  # eg. 14
    minor_version = g.inspect_get_minor_version(root)  # eg. 0
    hostname = g.inspect_get_hostname(root)  # string
    product_name = g.inspect_get_product_name(root)  # string
    product_variant = g.inspect_get_product_variant(root)  # string
    package_format = g.inspect_get_package_format(root)  # string

    # For inspect_list_applications and inspect_get_icon we
    # require that the guest filesystems are mounted.  However
    # don't fail if this is not possible (I'm looking at you,
    # FreeBSD).
    filesystems_mounted = False
    # Mount up the disks, like guestfish --ro -i.

    # Sort keys by length, shortest first, so that we end up
    # mounting the filesystems in the correct order.
    mps = g.inspect_get_mountpoints(root)

    mps = sorted(mps.items(), key=lambda k: len(k[0]))
    for mp, dev in mps:
        try:
            g.mount_ro(dev, mp)
            filesystems_mounted = True
        except Exception:
            log.exception("%s: exception mounting %s on %s (ignored)", prettyvm, dev, mp)

    icon = None
    apps = None
    if filesystems_mounted:
        # string containing PNG data
        icon = g.inspect_get_icon(root, favicon=0, highquality=1)
        if icon is None or len(icon) == 0:
            # no high quality icon, try a low quality one
            icon = g.inspect_get_icon(root, favicon=0, highquality=0)
            if icon is None or len(icon) == 0:
                icon = None

        # Inspection applications.
        try:
            gapps = g.inspect_list_applications2(root)
            # applications listing worked, so make apps a real list
            # (instead of None)
            apps = []
            for gapp in gapps:
                app = vmmInspectionApplication()
                if gapp["app2_name"]:
                    app.name = gapp["app2_name"]
                if gapp["app2_display_name"]:
                    app.display_name = gapp["app2_display_name"]
                app.epoch = gapp["app2_epoch"]
                if gapp["app2_version"]:
                    app.version = gapp["app2_version"]
                if gapp["app2_release"]:
                    app.release = gapp["app2_release"]
                if gapp["app2_summary"]:
                    app.summary = gapp["app2_summary"]
                if gapp["app2_description"]:
                    app.description = gapp["app2_description"]
                apps.append(app)
        except Exception:
            log.exception("%s: exception while listing apps (ignored)", prettyvm)

    # Force the libguestfs handle to close right now.
    del g

    # Log what we found.
    log.debug(
        "%s: detected operating system: %s %s %d.%d (%s) (%s)",
        prettyvm,
        os_type,
        distro,
        major_version,
        minor_version,
        product_name,
        package_format,
    )
    log.debug("hostname: %s", hostname)
    if icon:
        log.debug("icon: %d bytes", len(icon))
    if apps:
        log.debug("# apps: %d", len(apps))

    data = vmmInspectionData()
    data.os_type = str(os_type)
    data.distro = str(distro)
    data.major_version = int(major_version)
    data.minor_version = int(minor_version)
    data.hostname = str(hostname)
    data.product_name = str(product_name)
    data.product_variant = str(product_variant)
    data.icon = icon
    data.applications = list(apps or [])
    data.package_format = str(package_format)

    return data


class vmmInspection(vmmGObject):
    _libguestfs_installed = None

    @classmethod
    def get_instance(cls):
        if not cls._instance:
            if not cls.libguestfs_installed():
                return None  # pragma: no cover
            cls._instance = vmmInspection()
        return cls._instance

    @classmethod
    def libguestfs_installed(cls):
        if cls._libguestfs_installed is None:
            try:
                import guestfs as ignore  # pylint: disable=import-error

                log.debug("python guestfs is installed")
                cls._libguestfs_installed = True
            except ImportError:  # pragma: no cover
                log.debug("python guestfs is not installed")
                cls._libguestfs_installed = False
            except Exception:  # pragma: no cover
                log.debug("error importing guestfs", exc_info=True)
                cls._libguestfs_installed = False
        return cls._libguestfs_installed

    def __init__(self):
        vmmGObject.__init__(self)
        self._cleanup_on_app_close()

        self._thread = None

        self._q = queue.Queue()
        self._cached_data = {}
        self._uris = []

        val = self.config.get_libguestfs_inspect_vms()
        log.debug("libguestfs gsetting enabled=%s", str(val))
        if not val:
            return

        connmanager = vmmConnectionManager.get_instance()
        connmanager.connect("conn-added", self._conn_added_cb)
        connmanager.connect("conn-removed", self._conn_removed_cb)
        for conn in connmanager.conns.values():
            self._conn_added_cb(connmanager, conn)  # pragma: no cover

        self._start()

    def _cleanup(self):
        self._stop()
        self._q = queue.Queue()
        self._cached_data = {}

    def _conn_added_cb(self, connmanager, conn):
        uri = conn.get_uri()
        if uri in self._uris:
            return  # pragma: no cover

        self._uris.append(uri)
        conn.connect("vm-added", self._vm_added_cb)
        for vm in conn.list_vms():  # pragma: no cover
            self._vm_added_cb(conn, vm.get_name())

    def _conn_removed_cb(self, connmanager, uri):
        self._uris.remove(uri)

    def _vm_added_cb(self, conn, vm):
        # Called by the main thread whenever a VM is added to vmlist.
        name = vm.get_name()
        if name.startswith("guestfs-"):  # pragma: no cover
            log.debug("ignore libvirt/guestfs temporary VM %s", name)
            return

        self._q.put((conn.get_uri(), vm.get_name()))

    def _start(self):
        self._thread = threading.Thread(name="inspection thread", target=self._run)
        self._thread.daemon = True
        self._thread.start()

    def _stop(self):
        if self._thread is None:
            return

        self._q.put(None)
        self._thread = None

    def _run(self):
        # Process everything on the queue.  If the queue is empty when
        # called, block.
        while True:
            data = self._q.get()
            if data is None:
                log.debug("libguestfs queue vm=None, exiting thread")
                return
            uri, vmname = data
            self._process_vm(uri, vmname)
            self._q.task_done()

    def _process_vm(self, uri, vmname):
        connmanager = vmmConnectionManager.get_instance()
        conn = connmanager.conns.get(uri)
        if not conn:
            return  # pragma: no cover

        vm = conn.get_vm_by_name(vmname)
        if not vm:
            return  # pragma: no cover

        # Try processing a single VM, keeping into account whether it was
        # visited already, and whether there are cached data for it.
        def _set_vm_inspection_data(_data):
            vm.set_inspection_data(_data)
            self._cached_data[vm.get_uuid()] = _data

        prettyvm = conn.get_uri() + ":" + vm.get_name()
        vmuuid = vm.get_uuid()
        if vmuuid in self._cached_data:
            data = self._cached_data.get(vmuuid)
            if vm.inspection != data:
                log.debug("Found cached data for %s", prettyvm)
                _set_vm_inspection_data(data)
            return

        try:
            data = self._inspect_vm(conn, vm)
        except Exception as e:  # pragma: no cover
            data = _inspection_error(_("Error inspection VM: %s") % str(e))
            log.exception("%s: exception while processing", prettyvm)

        _set_vm_inspection_data(data)

    def _inspect_vm(self, conn, vm):
        if self._thread is None:
            return  # pragma: no cover

        if conn.is_remote():  # pragma: no cover
            return _inspection_error(_("Cannot inspect VM on remote connection"))
        if conn.is_test():
            return _make_fake_data(vm)

        return _perform_inspection(conn, vm)  # pragma: no cover

    ##############
    # Public API #
    ##############

    def vm_refresh(self, vm):
        log.debug("Refresh requested for vm=%s", vm.get_name())

        # When refreshing the inspection data of a VM,
        # all we need is to remove it from the "seen" cache,
        # as the data itself will be replaced once the new
        # results are available.
        self._cached_data.pop(vm.get_uuid(), None)
        self._q.put((vm.conn.get_uri(), vm.get_name()))



# Copyright (C) 2011, 2013 Red Hat, Inc.
# Copyright (C) 2011 Cole Robinson <crobinso@redhat.com>
#
# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

from virtinst import NodeDevice

from .libvirtobject import vmmLibvirtObject


def _usb_pretty_name(xmlobj):
    # product/vendor name fields can be messy, this tries to cope
    product = str(xmlobj.product_name or "").strip()
    vendor = str(xmlobj.vendor_name or "").strip()
    product = product or str(xmlobj.product_id or "")
    vendor = vendor or str(xmlobj.vendor_id or "")
    busstr = "%.3d:%.3d" % (int(xmlobj.bus), int(xmlobj.device))
    desc = "%s %s %s" % (busstr, vendor, product)
    return desc


def _pretty_name(xmlobj):
    if xmlobj.device_type == "net":
        return _("Interface %s") % xmlobj.interface or xmlobj.name

    if xmlobj.device_type == "pci":
        devstr = "%.4X:%.2X:%.2X:%X" % (
            int(xmlobj.domain),
            int(xmlobj.bus),
            int(xmlobj.slot),
            int(xmlobj.function),
        )
        return "%s %s %s" % (devstr, xmlobj.vendor_name, xmlobj.product_name)
    if xmlobj.device_type == "usb_device":
        return _usb_pretty_name(xmlobj)

    if xmlobj.device_type == "drm":
        parent = NodeDevice.lookupNodedevByName(xmlobj.conn, xmlobj.parent)

        # https://github.com/virt-manager/virt-manager/issues/328
        # Apparently we can't depend on successful parent lookup
        pretty_parent = xmlobj.parent
        if parent:
            pretty_parent = _pretty_name(parent)

        return "%s (%s)" % (pretty_parent, xmlobj.drm_type)

    return xmlobj.name  # pragma: no cover


class vmmNodeDevice(vmmLibvirtObject):
    def __init__(self, conn, backend, key):
        vmmLibvirtObject.__init__(self, conn, backend, key, NodeDevice)

    def _conn_tick_poll_param(self):
        return "pollnodedev"  # pragma: no cover

    def class_name(self):
        return "nodedev"

    def _XMLDesc(self, flags):
        return self._backend.XMLDesc(flags)

    def _using_events(self):
        return self.conn.using_node_device_events

    def _get_backend_status(self):
        is_active = True
        if self.conn.support.nodedev_isactive(self._backend):
            is_active = self._backend.isActive()  # pragma: no cover

        if self.conn.is_test() and self.xmlobj.name.endswith("-fakeinactive"):
            # Testsuite hack to mock inactive device
            is_active = False

        return is_active and self._STATUS_ACTIVE or self._STATUS_INACTIVE

    def pretty_name(self):
        return _pretty_name(self.xmlobj)
